from dataclasses import dataclass


@dataclass(frozen=True)
class CrawlItem:
    url: str
    depth: int


@dataclass(frozen=True)
class RetryPolicy:
    max_retries: int = 4
    timeout_s: float = 15.0
    backoff_base_s: float = 0.7
    backoff_jitter_s: float = 0.4
    delay_between_requests_s: float = 0.2



from typing import List

from bluer_options.terminal import show_usage, xtra


def help_build_corpus(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download,upload", mono=mono)

    args = []

    return show_usage(
        [
            "@rag",
            "build_corpus",
            f"[{options}]",
            "[.|<crawl-object-name>]",
            "[-|<corpus-object-name>]",
        ]
        + args,
        "build rag.",
        mono=mono,
    )


def help_query(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download", mono=mono)

    return show_usage(
        [
            "@rag",
            "query",
            f"[{options}]",
            "[.|<corpus-object-name>]",
            "<query>",
            "[-|<query-object-name>]",
        ],
        "query <query> on <corpus-object-name> -> <query-object-name>.",
        mono=mono,
    )


help_functions = {
    "build_corpus": help_build_corpus,
    "query": help_query,
}



from bluer_ai.help.generic import help_functions as generic_help_functions

from bluer_agent import ALIAS
from bluer_agent.help.assistant import help_assistant
from bluer_agent.help.audio import help_functions as help_audio
from bluer_agent.help.rag import help_functions as help_rag
from bluer_agent.help.chat import help_functions as help_chat
from bluer_agent.help.crawl import help_functions as help_crawl
from bluer_agent.help.transcribe import help as help_transcribe
from bluer_agent.help.voice import help_functions as help_voice

help_functions = generic_help_functions(plugin_name=ALIAS)

help_functions.update(
    {
        "assistant": help_assistant,
        "audio": help_audio,
        "chat": help_chat,
        "crawl": help_crawl,
        "rag": help_rag,
        "transcribe": help_transcribe,
        "voice": help_voice,
    }
)



import time
from typing import Tuple

from blueness import module
from bluer_options import string
from bluer_objects import file
from bluer_objects import objects
from bluer_objects.metadata import post_to_object
from bluer_objects.html_report import HTMLReport

from bluer_agent import NAME
from bluer_agent.assets.path import get_path
from bluer_agent.audio.play import play
from bluer_agent.audio.properties import AudioProperties
from bluer_agent.chat.functions import chat
from bluer_agent.chat.messages import List_of_Messages
from bluer_agent.rag.corpus.context import Context
from bluer_agent.rag.prompt.single_root import build_prompt
from bluer_agent.transcription.functions import transcribe
from bluer_agent.voice.functions import generate_voice
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)

greeting: str = "سلام، من رنگین هستم. چطور می‌تونم کمک‌تون کنم؟"


def converse(
    context: Context,
    object_name: str,
    greeting: str = greeting,
    language: str = "fa",
    audio_properties: AudioProperties = AudioProperties(),
    loop: bool = True,
    post_metadata: bool = True,
    verbose: bool = False,
) -> Tuple[bool, List_of_Messages]:
    logger.info(
        "{}.converse:context[{}] -{}{}> {}".format(
            NAME,
            context.object_name,
            "loop-" if loop else "",
            "metadata-" if post_metadata else "",
            object_name,
        )
    )

    audio_prompt: str = greeting

    list_of_messages = List_of_Messages()

    success: bool = True
    while True:
        html_report = HTMLReport(template=get_path("./query.html"))

        success, filename = generate_voice(
            object_name=object_name,
            sentence=audio_prompt,
        )
        if not success:
            break

        if not play(
            object_name=object_name,
            filename=filename,
        ):
            success = False
            break

        time.sleep(0.1)

        file.delete(
            objects.path_of(
                object_name=object_name,
                filename=filename,
            ),
            log=verbose,
        )

        time.sleep(0.9)

        filename = "{}.wav".format(string.timestamp())

        success, query = transcribe(
            object_name=object_name,
            filename=filename,
            language=language,
            record=True,
            properties=audio_properties,
        )
        if not success or not query:
            break

        success, query_context = context.generate(
            query=query,
            html_report=html_report,
        )
        if not success:
            break

        success, reply = chat(
            messages=build_prompt(
                query=query,
                context=query_context["chunks"],
                previous_messages=list_of_messages,
            ),
            html_report=html_report,
        )
        if not success:
            break

        success, reply_sentence = context.understand_reply(reply)
        if not success:
            break

        list_of_messages.messages.append(
            {
                "role": "assistant",
                "content": reply_sentence,
            }
        )

        if not html_report.save(
            object_name=object_name,
            filename=file.add_extension(filename, "html"),
        ):
            success = False
            break

        logger.info(
            "query: {}, reply: {}".format(
                string.pretty_bytes(len(query)),
                string.pretty_bytes(len(reply_sentence)),
            )
        )

        audio_prompt = reply_sentence

        if not loop:
            break

    if not post_to_object(
        object_name,
        "conversation",
        list_of_messages.messages,
    ):
        success = False

    return success, list_of_messages



from __future__ import annotations

import argparse

from blueness import module

from bluer_agent import env
from bluer_agent import NAME
from bluer_agent.assistant.endpoints import app
from bluer_agent.logger import logger

# needed, although seem unused :)
from bluer_agent.assistant.endpoints.add_step import add_step
from bluer_agent.assistant.endpoints.auto_generate_subject import auto_generate_subject
from bluer_agent.assistant.endpoints.delete_project import delete_project
from bluer_agent.assistant.endpoints.delete_requirement import delete_requirement
from bluer_agent.assistant.endpoints.delete_step import delete_step
from bluer_agent.assistant.endpoints.home import home
from bluer_agent.assistant.endpoints.move_step_down import move_step_down
from bluer_agent.assistant.endpoints.move_step_up import move_step_up
from bluer_agent.assistant.endpoints.next import next
from bluer_agent.assistant.endpoints.new import new
from bluer_agent.assistant.endpoints.open_project import open_project
from bluer_agent.assistant.endpoints.prev import prev
from bluer_agent.assistant.endpoints.refresh_list_of_projects import (
    refresh_list_of_projects,
)
from bluer_agent.assistant.endpoints.save import save
from bluer_agent.assistant.endpoints.save_objective import save_objective
from bluer_agent.assistant.endpoints.save_step import save_step
from bluer_agent.assistant.endpoints.submit import submit
from bluer_agent.assistant.endpoints.up import up

NAME = module.name(__file__, NAME)


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "--port",
    type=int,
    default=env.BLUER_AGENT_ASSISTANT_PORT,
)
parser.add_argument(
    "--hostname",
    type=str,
    default="0.0.0.0",
)
parser.add_argument(
    "--object_name",
    type=str,
)
args = parser.parse_args()

logger.info(
    "{} on host:{}, port:{} -> {}".format(
        NAME,
        args.hostname,
        args.port,
        args.object_name,
    )
)

# if provided, home (/) will redirect to this.
if args.object_name:
    app.config["object_name"] = args.object_name

app.run(
    debug=True,
    host=args.hostname,
    port=args.port,
)



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_agent import env
from bluer_agent.transcription.functions import list_of_languages

audio_properties_args = [
    "[--channels <1>]",
    "[--crop_silence <0>]",
    "[--length <30>]",
    "[--rate <16000>]",
]


def help_converse(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("~download,upload", mono=mono)

    args = sorted(
        [
            "[--greeting <str>]",
            "[--language {}]".format(" | ".join(list_of_languages)),
        ]
        + audio_properties_args
    )

    return show_usage(
        [
            "@audio",
            "converse",
            f"[{options}]",
            "[{} | <context-object-name>]".format(
                env.BLUER_AGENT_RAG_CORPUS_SINGLE_ROOT_TEST_OBJECT
            ),
            "[-|<object-name>]",
        ]
        + args,
        "have a conversation based on the context.",
        mono=mono,
    )


def help_install(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@audio",
            "install",
        ],
        "install audio.",
        mono=mono,
    )


def help_play(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download,filename=<audio.wav>", mono=mono)

    return show_usage(
        [
            "@audio",
            "play",
            f"[{options}]",
            "[.|<object-name>]",
        ],
        "play <object-name>/<audio.wav>.",
        mono=mono,
    )


def help_record(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "".join(
        [
            xtra("filename=<audio.wav>,", mono=mono),
            "play",
            xtra(",upload", mono=mono),
        ]
    )

    return show_usage(
        [
            "@audio",
            "record",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + audio_properties_args,
        "record <object-name>/<audio.wav>.",
        mono=mono,
    )


def help_test(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@audio",
            "test",
        ],
        "test audio.",
        mono=mono,
    )


help_functions = {
    "converse": help_converse,
    "install": help_install,
    "play": help_play,
    "record": help_record,
    "test": help_test,
}



from __future__ import annotations

import gzip
import json
from typing import Callable, List
import requests
import numpy as np
from tqdm import tqdm

from blueness import module
from bluer_options import string
from bluer_options.logger.config import shorten_text
from bluer_objects import objects, file

from bluer_agent import NAME
from bluer_agent import env
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def _read_jsonl_gz(filename: str):
    with gzip.open(filename, "rt", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            yield json.loads(line)


def _read_json_gz(filename: str):
    with gzip.open(filename, "rt", encoding="utf-8") as f:
        return json.loads(f.read())


import requests


def embed_fn(texts):
    texts = [line[:8192] for line in texts]

    logger.info("embed_fn({})".format(shorten_text(" ".join(texts))))

    r = requests.post(
        f"{env.BLUER_AGENT_EMBEDDING_ENDPOINT}/embeddings",
        headers={
            "Authorization": f"apikey {env.BLUER_AGENT_API_KEY}",
            "Content-Type": "application/json",
        },
        json={
            "model": env.BLUER_AGENT_EMBEDDING_MODEL_NAME,
            "input": texts,
        },
        timeout=60,
    )
    r.raise_for_status()

    data = r.json()
    return [item["embedding"] for item in data["data"]]


def embed(
    object_name: str,
    embed_fn: Callable[[List[str]], List[List[float]]],
    batch_size: int = 64,
) -> bool:
    logger.info(f"{NAME}.embed: {object_name}")

    corpus_filename = objects.path_of(
        object_name=object_name,
        filename="corpus.jsonl.gz",
    )
    roots_filename = objects.path_of(
        object_name=object_name,
        filename="roots.json.gz",
    )

    chunks_vec_filename = objects.path_of(
        object_name=object_name,
        filename="corpus.embeddings.npy",
    )
    chunks_meta_filename = objects.path_of(
        object_name=object_name,
        filename="corpus.meta.jsonl.gz",
    )

    roots_vec_filename = objects.path_of(
        object_name=object_name,
        filename="roots.embeddings.npy",
    )
    roots_meta_filename = objects.path_of(
        object_name=object_name,
        filename="roots.meta.json.gz",
    )

    # --- embed roots (for "most relevant root" selection) ---
    roots = _read_json_gz(roots_filename)
    root_names = sorted(list(roots.keys()))
    root_texts = [roots[name].get("text", "") for name in root_names]

    root_vectors = []
    for i in tqdm(range(0, len(root_texts), batch_size), desc="embedding roots"):
        batch = root_texts[i : i + batch_size]
        root_vectors.extend(embed_fn(batch))

    root_vectors = np.asarray(root_vectors, dtype=np.float32)
    np.save(roots_vec_filename, root_vectors)

    roots_meta = {
        "roots": root_names,
        "shape": list(root_vectors.shape),
        "source": roots_filename,
    }
    with gzip.open(roots_meta_filename, "wt", encoding="utf-8") as f:
        f.write(json.dumps(roots_meta, ensure_ascii=False))

    logger.info(
        "roots vec -{}-> {}".format(
            string.pretty_bytes(file.size(roots_vec_filename)),
            roots_vec_filename,
        )
    )
    logger.info(
        "roots meta -{}-> {}".format(
            string.pretty_bytes(file.size(roots_meta_filename)),
            roots_meta_filename,
        )
    )

    # --- embed chunks (for evidence extraction within selected root) ---
    vectors = []
    batch_texts: List[str] = []
    batch_meta: List[dict] = []

    with gzip.open(chunks_meta_filename, "wt", encoding="utf-8") as meta_f:
        for record in tqdm(_read_jsonl_gz(corpus_filename), desc="embedding corpus"):
            text = record.get("text", "")
            if not text:
                continue

            batch_texts.append(text)
            batch_meta.append(
                {
                    "root": record.get("root", ""),
                    "url": record.get("url", ""),
                    "chunk_id": record.get("chunk_id", -1),
                }
            )

            if len(batch_texts) >= batch_size:
                batch_vecs = embed_fn(batch_texts)
                vectors.extend(batch_vecs)

                for m in batch_meta:
                    meta_f.write(json.dumps(m, ensure_ascii=False) + "\n")

                batch_texts = []
                batch_meta = []

        if batch_texts:
            batch_vecs = embed_fn(batch_texts)
            vectors.extend(batch_vecs)

            for m in batch_meta:
                meta_f.write(json.dumps(m, ensure_ascii=False) + "\n")
    logger.info(
        "chunk meta -{}-> {}".format(
            string.pretty_bytes(file.size(chunks_meta_filename)),
            chunks_meta_filename,
        )
    )

    vectors = np.asarray(vectors, dtype=np.float32)
    np.save(chunks_vec_filename, vectors)
    logger.info(
        "chunks vec -{}-{}-> {}".format(
            string.pretty_shape_of_matrix(vectors),
            string.pretty_bytes(file.size(chunks_vec_filename)),
            chunks_vec_filename,
        )
    )

    return True



from typing import Tuple, Dict, List
import gzip
import json
import numpy as np
from functools import reduce

from blueness import module
from bluer_options.logger.config import log_list
from bluer_objects import file, storage
from bluer_objects import objects
from bluer_objects.html_report import HTMLReport
from bluer_objects.storage.policies import DownloadPolicy

from bluer_agent import NAME
from bluer_agent.host import signature
from bluer_agent.rag.corpus.embed import embed_fn
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)


def _cosine(a: np.ndarray, b: np.ndarray) -> float:
    return float(np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b)))


class Context:
    def __init__(
        self,
        object_name: str,
        download: bool = False,
    ):
        self.list_of_roots: List[str] = []

        self.object_name = object_name

        if download:
            storage.download(
                self.object_name,
                policy=DownloadPolicy.DOESNT_EXIST,
            )

        # --- load corpus embeddings ---
        self.corpus_vec = np.load(
            objects.path_of(
                object_name=object_name,
                filename="corpus.embeddings.npy",
            )
        )

        self.corpus_meta_file = objects.path_of(
            object_name=self.object_name,
            filename="corpus.meta.jsonl.gz",
        )

        self.corpus_text_file = objects.path_of(
            object_name=self.object_name,
            filename="corpus.jsonl.gz",
        )

        logger.info(
            "{} created for {}.".format(
                self.__class__.__name__,
                self.object_name,
            )
        )

    def generate(
        self,
        query: str,
        top_k: int = 5,
        html_report: HTMLReport = HTMLReport(),
    ) -> Tuple[bool, Dict]:
        logger.info(
            '{}[{}].generate("{}")'.format(
                self.__class__.__name__,
                self.object_name,
                query,
            )
        )

        # --- embed query ---
        q_vec = np.asarray(embed_fn([query])[0], dtype=np.float32)

        candidates: List[Tuple[float, dict]] = []

        self.list_of_roots = []
        with gzip.open(
            self.corpus_meta_file,
            "rt",
            encoding="utf-8",
        ) as meta_f, gzip.open(
            self.corpus_text_file,
            "rt",
            encoding="utf-8",
        ) as text_f:
            for i, (meta_line, text_line) in enumerate(zip(meta_f, text_f)):
                meta = json.loads(meta_line)

                if meta["root"] not in self.list_of_roots:
                    self.list_of_roots.append(meta["root"])

                record = json.loads(text_line)
                score = _cosine(q_vec, self.corpus_vec[i])

                candidates.append(
                    (
                        score,
                        {
                            "root": meta["root"],
                            "url": meta["url"],
                            "chunk_id": meta["chunk_id"],
                            "text": record["text"],
                        },
                    )
                )

        candidates.sort(key=lambda x: x[0], reverse=True)
        top = candidates[:top_k]

        log_list(logger, "reviewed", self.list_of_roots, "root(s)")

        chunks = [
            {
                **item,
                "score": round(score, 4),
            }
            for score, item in top
        ]

        for chunk in chunks:
            logger.info(
                "{} #{} @ {:.2f}: {}".format(
                    chunk["root"],
                    chunk["chunk_id"],
                    chunk["score"],
                    chunk["text"][:300],
                )
            )

        context = {
            "chunks": chunks,
        }

        # TODO: change to HTMLReport - 2026-02-05 - nice to have.
        file.save_text(
            objects.path_of(
                object_name=self.object_name,
                filename="result.txt",
            ),
            [
                f"query: {query}",
                "",
                "retrievals:",
            ]
            + [
                "{} #{} @ {:.2f}: {}".format(
                    chunk["root"],
                    chunk["chunk_id"],
                    chunk["score"],
                    chunk["text"][:300],
                )
                for chunk in chunks
            ],
            log=True,
        )

        html_report.replace(
            {
                "context_count:::": str(len(context["chunks"])),
                "corpus_name:::": self.object_name,
                "query:::": query,
                "query_dir:::": "ltr",
                "query_lang:::": "fa",
                "host_signature:::": " | ".join(signature()),
                "text_dir:::": "ltr",
                "text_lang:::": "fa",
                "title:::": "rag query output",
            }
        ).replace(
            {
                "context:::": reduce(
                    lambda x, y: x + y,
                    [
                        [
                            "<tr>",
                            '    <td dir="ltr">',
                            '        <a href="{}" target="_blank" rel="noreferrer" title="{}">{}#{}</a>'.format(
                                chunk["url"],
                                chunk["text"],
                                chunk["root"],
                                chunk["chunk_id"],
                            ),
                            "    </td>",
                            '    <td class="mono">{:.2f}</td>'.format(chunk["score"]),
                            "</tr>",
                            "",
                        ]
                        for chunk in context["chunks"]
                    ],
                    [],
                ),
            },
            contains=True,
        )

        return True, context

    def understand_reply(self, reply: str) -> Tuple[bool, str]:
        if len(self.list_of_roots) == 1:
            return True, reply

        try:
            reply_dict = json.loads(reply)
        except Exception as e:
            logger.error(e)
            return False, ""

        for field in [
            "winner_root",
            "argument_fa",
        ]:
            if field not in reply_dict:
                logger.error(f"{field} not found.")
                return False, ""

        logger.info("🥇 winner is {}!".format(reply_dict["winner_root"]))

        return True, reply_dict["argument_fa"]



from typing import List
import argparse

from bluer_options import string
from bluer_options.host import is_rpi

from bluer_agent.audio import env


class AudioProperties:
    def __init__(
        self,
        channels: int = 1,
        crop_silence: bool = True,
        length: int = 30,  # in seconds
        rate: int = 48000,
    ):
        self.channels: int = channels
        self.crop_silence: bool = crop_silence
        self.length: int = length
        self.rate: int = rate

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        parser.add_argument(
            "--channels",
            type=int,
            default=1,
        )
        parser.add_argument(
            "--crop_silence",
            type=int,
            default=1,
            help="0 | 1",
        )
        parser.add_argument(
            "--length",
            type=int,
            default=30,
            help="in seconds",
        )
        parser.add_argument(
            "--rate",
            type=int,
            default=48000,
        )

    def as_str(self):
        return "length<{}, sample rate:{}, {} channel(s){}".format(
            string.pretty_duration(self.length),
            self.rate,
            self.channels,
            ", crop silence" if self.crop_silence else "",
        )

    def record_command(
        self,
        filename: str,
    ) -> List[str]:
        # input device:
        # - On Raspberry Pi with mic as default: don't specify device
        # - If you need a specific ALSA device: add `-t alsa hw:1,0` after rec
        return (
            (
                [
                    "sudo",
                    "-u",
                    "pi",
                    "-H",
                ]
                if is_rpi()
                else []
            )
            + [
                "sox" if is_rpi() else "rec",
                "-V1",
            ]
            + (
                [
                    "-t",
                    "alsa",
                    '"plughw:CARD=C930e,DEV=0"',
                ]
                if is_rpi()
                else []
            )
            + [
                f'-r "{self.rate}"',
                f'-c "{self.channels}"',
                filename,
                f"trim 0 {self.length}",
            ]
            + (
                [
                    "silence",
                    f'1 "{env.BLUER_AGENT_AUDIO_LISTEN_START_HOLD}" "{env.BLUER_AGENT_AUDIO_LISTEN_START_THRESHOLD}"',
                    f'1 "{env.BLUER_AGENT_AUDIO_LISTEN_STOP_HOLD}" "{env.BLUER_AGENT_AUDIO_LISTEN_STOP_THRESHOLD}"',
                ]
                if self.crop_silence
                else []
            )
        )



from typing import List, Dict, Tuple
import requests
import re
from functools import reduce

from blueness import module
from bluer_options.logger.config import log_list
from bluer_objects.html_report import HTMLReport

from bluer_agent import NAME
from bluer_agent import env
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def chat(
    messages: List[Dict],
    remove_thoughts: bool = True,
    html_report: HTMLReport = HTMLReport(),
) -> Tuple[bool, str]:
    log_list(
        logger,
        "{}.chat({}):".format(
            NAME,
            env.BLUER_AGENT_CHAT_MODEL_NAME,
        ),
        [str(message) for message in messages],
        "message(s)",
        max_count=-1,
    )

    html_report.replace(
        {
            "message_count:::": str(len(messages)),
            "model_name:::": "{} @ temperature={}".format(
                env.BLUER_AGENT_CHAT_MODEL_NAME,
                env.BLUER_AGENT_CHAT_TEMPERATURE,
            ),
        }
    ).replace(
        {
            "messages:::": reduce(
                lambda x, y: x + y,
                [
                    [
                        '<div class="card" style="margin-bottom:12px;">',
                        '    <details class="section">',
                        '        <summary class="header">',
                        '            <div class="header-left">',
                        '                <span class="chev">›</span>',
                        '                <h2 class="mono" style="font-size:14px;">{}</h2>'.format(
                            message["role"],
                        ),
                        "            </div>",
                        f'            <div class="pill mono">#{index+1}</div>',
                        "        </summary>",
                        "",
                        '        <div class="content">',
                        '            <div class="block">',
                        '                <pre dir="rtl" lang="fa">',
                        message["content"],
                        "                </pre>",
                        "            </div>",
                        "        </div>",
                        "    </details>",
                        "</div>",
                    ]
                    for index, message in enumerate(messages)
                ],
                [],
            ),
        },
        contains=True,
    )

    headers = {
        "Authorization": f"apikey {env.BLUER_AGENT_API_KEY}",
        "Content-Type": "application/json",
    }

    payload = {
        "model": env.BLUER_AGENT_CHAT_MODEL_NAME,
        "messages": messages,
        "max_tokens": env.BLUER_AGENT_CHAT_MAX_TOKENS,
        "temperature": env.BLUER_AGENT_CHAT_TEMPERATURE,
    }

    try:
        response = requests.post(
            "{}/chat/completions".format(env.BLUER_AGENT_CHAT_ENDPOINT),
            headers=headers,
            json=payload,
            timeout=env.BLUER_AGENT_CHAT_TIMEOUT,
        )
    except Exception as e:
        logger.error(f"failed to send request: {e}")
        return False, ""

    success = True
    response_json = {}
    if response.status_code // 100 != 2:  # Check if status code is not in the 2xx range
        logger.info(
            "failed, status_code: {}, reason: {}.".format(
                response.status_code,
                response.reason,
            )
        )
        success = False
    else:
        try:
            response_json = response.json()
        except Exception as e:
            logger.error(f"failed to parse response to json: {e}, response: {response}")
            success = False

    if success:
        if not isinstance(response_json, dict):
            logger.error("response is not a dict")
            success = False
        elif "choices" not in response_json:
            logger.error("choices not in response")
            success = False
        elif len(response_json["choices"]) == 0:
            logger.error("response.choices is empty")
            success = False
        elif len(response_json["choices"]) > 1:
            logger.warning(
                "{} choice(s), will use the first one.".format(response_json["choices"])
            )
    if not success:
        return success, ""

    text = response_json["choices"][0].get("message", {}).get("content", "")

    html_report.replace(
        {
            "thoughts:::": [text],
        },
        contains=True,
    )

    if remove_thoughts and text:
        text = re.sub(
            r"<think>.*?</think>",
            "",
            text,
            flags=re.DOTALL | re.IGNORECASE,
        )

    text = re.sub(r"\s+", " ", text).strip()

    logger.info(text)

    html_report.replace(
        {
            "reply:::": [text],
        },
        contains=True,
    )

    return success, text



docs = [
    {
        "path": f"../docs/rag{suffix}",
    }
    for suffix in [
        "",
        "/build/",
        "/build/one.md",
        "/build/hnpagency.md",
        "/build/badkoobeh.md",
        "/query/",
        "/query/one.md",
        "/query/hnpagency.md",
        "/query/badkoobeh.md",
    ]
]



from typing import List

from bluer_options.terminal import show_usage


def help_(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "upload"

    args = [
        "[--filename <report.html>]",
        "[--prompt <prompt>]",
        "[--remove_thoughts 1]",
    ]

    return show_usage(
        [
            "@chat",
            f"[{options}]",
            "[-|<object-name>]",
        ]
        + args,
        "validate chatting.",
        mono=mono,
    )


help_functions = {
    "": help_,
}



import argparse
import base64

from blueness import module
from blueness.argparse.generic import sys_exit
from bluer_objects import file
from bluer_objects import objects
from bluer_objects.html_report import HTMLReport

from bluer_agent import NAME
from bluer_agent.rag.query import query
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="query",
)
parser.add_argument(
    "--corpus_object_name",
    type=str,
)
parser.add_argument(
    "--query_object_name",
    type=str,
)
parser.add_argument(
    "--encoded_query",
    type=str,
)
parser.add_argument(
    "--top_k",
    type=int,
    default=5,
)
parser.add_argument(
    "--filename",
    type=str,
    default="query.html",
)
parser.add_argument(
    "--generate_html",
    type=int,
    default=1,
    help="0 | 1",
)
args = parser.parse_args()

success = False
if args.task == "query":
    html_report = HTMLReport(
        template=(
            file.absolute(
                "../assets/query.html",
                file.path(__file__),
            )
            if args.generate_html == 1
            else ""
        ),
    )

    success, _ = query(
        corpus_object_name=args.corpus_object_name,
        query=base64.b64decode(args.encoded_query).decode("utf-8"),
        top_k=args.top_k,
        html_report=html_report,
    )

    if success:
        success = html_report.save(
            object_name=args.query_object_name,
            filename=args.filename,
        )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from __future__ import annotations

import gzip
import json
import re

from tqdm import tqdm

from blueness import module
from bluer_options import string
from bluer_options.logger.config import log_list
from bluer_objects import objects
from bluer_objects import file as file_
from bluer_objects.metadata import post_to_object

from bluer_agent import NAME
from bluer_agent.crawl import file
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def _root_name(filename: str) -> str:
    name = file_.name(filename)
    for suffix in [".pkl.gz", ".pkl"]:
        if name.endswith(suffix):
            name = name[: -len(suffix)]
            break
    return name


def _normalize(text: str) -> str:
    text = text.replace("\u064a", "\u06cc")  # ي -> ی
    text = text.replace("\u0643", "\u06a9")  # ك -> ک
    text = text.replace("\u200c", " ")
    text = text.replace("\u200e", " ")
    text = text.replace("\u200f", " ")
    text = text.replace("\xad", " ")
    text = re.sub(r"\s+", " ", text).strip()
    return text


def _chunk(text: str, max_chars: int = 900, overlap: int = 120):
    text = _normalize(text)
    if not text:
        return
    step = max(1, max_chars - max(0, overlap))
    i = 0
    while i < len(text):
        chunk = text[i : i + max_chars].strip()
        if chunk:
            yield chunk
        i += step


def build(
    crawl_object_name: str,
    corpus_object_name: str,
) -> bool:
    logger.info(f"{NAME}.build: {crawl_object_name} -> {corpus_object_name}")

    list_of_filenames = [
        file_.name_and_extension(filename)
        for filename in file_.list_of(
            objects.path_of(
                object_name=crawl_object_name,
                filename="*.pkl.gz",
            )
        )
    ]
    log_list(
        logger,
        "processing",
        list_of_filenames,
        "file(s)",
    )

    corpus_filename = objects.path_of(
        object_name=corpus_object_name,
        filename="corpus.jsonl.gz",
    )
    roots_filename = objects.path_of(
        object_name=corpus_object_name,
        filename="roots.json.gz",
    )

    roots = {}

    with gzip.open(corpus_filename, "wt", encoding="utf-8") as f:
        for filename in tqdm(list_of_filenames):
            root = _root_name(filename)
            logger.info(
                "processing {} -> {} ...".format(
                    filename,
                    root,
                )
            )

            success, crawl = file.load(
                object_name=crawl_object_name,
                filename=filename,
            )
            if not success:
                return False

            agg_parts = []
            page_count = 0
            chunk_count = 0

            for url, text in tqdm(crawl.items()):
                page_count += 1
                text = _normalize(text)
                if text:
                    agg_parts.append(text[:2000])

                for chunk_id, chunk in enumerate(_chunk(text)):
                    record = {
                        "root": root,
                        "url": url,
                        "chunk_id": chunk_id,
                        "text": chunk,
                    }
                    f.write(json.dumps(record, ensure_ascii=False) + "\n")
                    chunk_count += 1

            roots[root] = {
                "root": root,
                "pages": page_count,
                "chunks": chunk_count,
                "text": _normalize(" ".join(agg_parts))[:200_000],
            }

    logger.info(
        "corpus -{}-> {}".format(
            string.pretty_bytes(file_.size(corpus_filename)),
            corpus_filename,
        )
    )

    with gzip.open(roots_filename, "wt", encoding="utf-8") as f:
        f.write(json.dumps(roots, ensure_ascii=False))
    logger.info(
        "roots -{}-> {}".format(
            string.pretty_bytes(file_.size(roots_filename)),
            roots_filename,
        ),
    )

    for root_info in roots.values():
        del root_info["text"]

    return post_to_object(
        corpus_object_name,
        "crawl",
        {
            "roots": roots,
            "source": crawl_object_name,
        },
    )



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_agent import NAME
from bluer_agent.voice.functions import (
    generate_voice,
    list_of_speakers,
    default_speaker,
)
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="generate",
)
parser.add_argument(
    "--sentence",
    type=str,
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--filename",
    type=str,
)
parser.add_argument(
    "--download",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--speaker",
    type=str,
    default=default_speaker,
    help=" | ".join(list_of_speakers),
)
parser.add_argument(
    "--speed",
    type=float,
    default=1.0,
    help=">0",
)
parser.add_argument(
    "--timestamp",
    type=int,
    default=0,
    help="0 | 1",
)
args = parser.parse_args()

success = False
if args.task == "generate":
    success, _ = generate_voice(
        object_name=args.object_name,
        filename=args.filename,
        sentence=args.sentence,
        download=args.download == 1,
        speaker=args.speaker,
        speed=args.speed,
        timestamp=args.timestamp == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_agent.voice.functions import list_of_speakers


def help_generate(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("~download,~play", mono=mono)

    args = [
        "[--speaker {}]".format(" | ".join(list_of_speakers)),
        "[--speed < >0 >]",
        "[--timestamp 1]",
    ]

    return show_usage(
        [
            "@voice",
            "generate",
            f"[{options}]",
            "[-|<object-name>]",
            '"<sentence>"',
        ]
        + args,
        "generate voice.",
        mono=mono,
    )


help_functions = {
    "generate": help_generate,
}



import gzip
import pickle
from typing import Dict, Tuple, List
from functools import reduce

from blueness import module
from bluer_options import string
from bluer_options.logger.config import log_list, shorten_text
from bluer_objects import file, path
from bluer_objects.html_report import HTMLReport
from bluer_objects import objects

from bluer_agent import NAME, ICON
from bluer_agent.host import signature
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)

"""
using gzip-compressed pickle:
    - binary
    - compact
    - easy to load back in Python
"""


def export(
    results: Dict[str, str],
    object_name: str,
    filename: str,
) -> bool:
    filename = file.add_extension(filename, "html")

    if not (
        HTMLReport(
            file.absolute(
                "../assets/review.html",
                file.path(__file__),
            ),
        )
        .replace(
            {
                "title:::": object_name,
                "signature:::": "{} {}".format(
                    ICON,
                    " | ".join(signature()),
                ),
            }
        )
        .replace(
            {
                "content:::": reduce(
                    lambda x, y: x + y,
                    [
                        [
                            "<details>",
                            f'    <summary><a href="{key}">{key}</a></summary>',
                            '    <div class="value">{}</div>'.format(
                                value.replace("\n", " ")
                            ),
                            "</details>",
                        ]
                        for key, value in results.items()
                    ],
                    [],
                ),
            },
            contains=True,
        )
        .save(
            object_name=object_name,
            filename=filename,
        )
    ):
        return False

    full_filename = objects.path_of(
        object_name=object_name,
        filename=filename,
    )

    logger.info(
        "{}.export: {} page(s) -> {} [{}]".format(
            NAME,
            len(results),
            full_filename,
            string.pretty_bytes(file.size(full_filename)),
        )
    )

    return True


def load(
    object_name: str,
    filename: str,
) -> Tuple[bool, Dict[str, str]]:
    full_filename = objects.path_of(
        object_name=object_name,
        filename=filename,
    )

    try:
        with gzip.open(full_filename, "rb") as f:
            payload = pickle.load(f)
    except Exception as e:
        logger.error(e)
        return False, {}

    if (
        not isinstance(payload, dict)
        or payload.get("format") != "site_text_collector"
        or "results" not in payload
    ):
        logger.error("unrecognized binary format.")
        return False, {}

    results = payload["results"]

    log_list(
        logger,
        "loaded from {}/{} [{}]".format(
            object_name,
            filename,
            string.pretty_bytes(file.size(full_filename)),
        ),
        [
            "{}: {}".format(
                key,
                shorten_text(text.replace("\n", " ")),
            )
            for key, text in results.items()
        ],
        "page(s)",
    )

    return True, results


def save(
    results: Dict[str, str],
    object_name: str,
    filename: str,
) -> bool:
    if not results:
        logger.warning("no pages collected; nothing to save.")
        return True

    full_filename = objects.path_of(
        object_name=object_name,
        filename=filename,
    )

    payload = {
        "format": "site_text_collector",
        "version": 1,
        "results": results,
    }
    try:
        with gzip.open(full_filename, "wb") as f:
            pickle.dump(payload, f, protocol=pickle.HIGHEST_PROTOCOL)
    except Exception as e:
        logger.error(e)
        return False

    logger.info(
        "{}.save: {} page(s) -> {}/{} [{}]".format(
            NAME,
            len(results),
            object_name,
            filename,
            string.pretty_bytes(file.size(full_filename)),
        )
    )
    return True



import argparse
from tqdm import tqdm

from blueness import module
from blueness.argparse.generic import sys_exit
from bluer_options.logger.config import log_list
from bluer_objects.metadata import get_from_object

from bluer_agent import NAME
from bluer_agent.crawl import file
from bluer_agent.crawl.collect import collect, CollectionProperties
from bluer_agent.crawl.review import review
from bluer_agent.logger import logger

NAME = module.name(__file__, NAME)

properties = CollectionProperties()

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="collect | review",
)
parser.add_argument(
    "--root",
    required=True,
    help="https://badkoobeh.com/",
)
parser.add_argument(
    "--page-count",
    type=int,
    default=25,
)
parser.add_argument(
    "--max-depth",
    type=int,
    default=2,
)
parser.add_argument(
    "--object_name",
    type=str,
    default="",
)
parser.add_argument(
    "--out",
    default="site_text.pkl.gz",
)
parser.add_argument(
    "--timeout",
    type=float,
    default=properties.timeout,
)
parser.add_argument(
    "--max-retries",
    type=int,
    default=properties.max_retries,
)
parser.add_argument(
    "--backoff-base",
    type=float,
    default=properties.backoff_base,
)
parser.add_argument(
    "--backoff-jitter",
    type=float,
    default=properties.backoff_jitter,
)
parser.add_argument(
    "--delay",
    type=float,
    default=properties.delay,
)
args = parser.parse_args()

success = False
if args.task == "collect":
    properties = CollectionProperties(
        timeout=args.timeout,
        max_retries=args.max_retries,
        backoff_base=args.backoff_base,
        backoff_jitter=args.backoff_jitter,
        delay=args.delay,
    )

    if args.root == "all":
        corpus_roots = get_from_object(args.object_name, "corpus", [])
        log_list(logger, "loaded", corpus_roots, "root(s)")

        success = True
        for root in tqdm(corpus_roots):
            if not collect(
                root=root,
                page_count=args.page_count,
                max_depth=args.max_depth,
                object_name=args.object_name,
                out="auto",
                properties=properties,
            ):
                success = False
                break
    else:
        success = collect(
            root=args.root,
            page_count=args.page_count,
            max_depth=args.max_depth,
            object_name=args.object_name,
            out=args.out,
            properties=properties,
        )
elif args.task == "review":
    success = review(
        object_name=args.object_name,
        root=args.root,
    )
else:
    success = None
sys_exit(logger, NAME, args.task, success)



from bluer_objects.README.alias import list_of_aliases

from bluer_agent import NAME
from bluer_agent.README import (
    aliases,
    assistant,
    audio,
    chat,
    crawl,
    rag,
    transcription,
    voice,
)

docs = (
    [
        {
            "path": "../..",
            "macros": {
                "aliases:::": list_of_aliases(NAME),
            },
        },
        {
            "path": "../docs",
        },
        {
            "path": "../docs/other.md",
        },
    ]
    + aliases.docs
    + assistant.docs
    + audio.docs
    + chat.docs
    + crawl.docs
    + rag.docs
    + transcription.docs
    + voice.docs
)



from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets_url


def image_template(suffix):
    return assets_url(
        suffix=f"bluer-agent/assistant/{suffix}",
        volume=3,
        blob=True,
    )


docs = [
    {
        "path": "../docs/assistant",
        "items": ImageItems(
            {
                image_template("screenshot-2026-02-13-15-15-06.png"): "",
                image_template("screenshot-2026-02-11-01-09-15.png"): "",
            }
        ),
    },
    {
        "path": "../docs/assistant/data-model.md",
    },
]



from blueness import module
from bluer_options import string
from bluer_options.host import is_rpi
from bluer_objects import file, objects
from bluer_objects.host import shell

from bluer_agent import NAME
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def play(
    object_name: str,
    filename: str = "audio.wav",
) -> bool:
    full_filename = objects.path_of(
        object_name=object_name,
        filename=filename,
    )
    if not file.exists(full_filename):
        logger.error(f"file not found: {full_filename}")
        return False

    logger.info(
        "{}.play: {}/{} ({})".format(
            NAME,
            object_name,
            filename,
            string.pretty_bytes(file.size(full_filename)),
        )
    )

    command = (
        (
            "ffmpeg -i {} -f wav - | aplay"
            if file.extension(filename) == "mp3"
            else "aplay {}"
        )
        if is_rpi()
        else "afplay {}"
    )

    return shell(
        command=command.format(full_filename),
        log=True,
    )


""



from flask import flash as flash_

from bluer_agent.logger import logger


def flash(
    message: str,
    category: str = "warning",
):
    icon: str = ""
    if category == "warning":
        logger.warning(message)
        icon = "⚠️ "
    elif category == "info":
        logger.info(message)
        icon = "ℹ️ "
    else:
        logger.info(f"{message} - category: {category}")

    return flash_(
        (f"{icon}{message}"),
        category,
    )



from typing import Tuple

from blueness import module
from bluer_objects.html_report import HTMLReport

from bluer_agent import NAME
from bluer_agent.rag.corpus.context import Context
from bluer_agent.rag.prompt import multi_root, single_root
from bluer_agent.chat.functions import chat
from bluer_agent.logger import logger


NAME = module.name(__file__, NAME)


def query(
    corpus_object_name: str,
    query: str,
    top_k: int = 5,
    html_report: HTMLReport = HTMLReport(),
) -> Tuple[bool, str]:
    logger.info(
        '{}.query[{}]("{}")'.format(
            NAME,
            corpus_object_name,
            query,
        )
    )

    context = Context(corpus_object_name)

    success, query_context = context.generate(
        query=query,
        top_k=top_k,
        html_report=html_report,
    )
    if not success:
        return success, ""

    return chat(
        messages=(
            single_root.build_prompt
            if len(context.list_of_roots) == 1
            else multi_root.build_prompt
        )(
            query=query,
            context=query_context["chunks"],
        ),
        html_report=html_report,
    )
